--[[
This script contains several useful functions for many kinds of collision detection.
It uses 2 basic data types, 'line' and 'sprite'.
'line' must be comprised of x1, y1, x2, y2 as non-nil numerical values.
'sprite' must be comprised of x, y, image where x and y are numerical and image is the love image class.
Note: a collisionMap key is needed for pixel-perfect collision. Pixel-perfect collision should be used as
little as possible because of its large memory consumption and possible slow downs.
The collision map may be constructed with the NewCollisionMap function with the associated filepath passed as argument.
--]]

--[[
You never know how someone wants to store their lines so I have provided a neutral line class constructor.
--]]
function NewLine(sx,sy,ex,ey)
	return {x1=sx,y1=sy,x2=ex,y2=ey}
end

function BoundingCircleCollide(S1,S2)
	local Rad1 = math.sqrt((S1.image:getWidth()/2)^2+(S1.image:getHeight()/2)^2)
	local Rad2 = math.sqrt((S2.image:getWidth()/2)^2+(S2.image:getHeight()/2)^2)
	local Dist = math.sqrt((S2.x-S1.x)^2+(S2.y-S1.y)^2)
	return Rad1+Rad2 > Dist
end

function BoundingCircleCollideRadius(S1,S2,R1,R2)
	local Dist = math.sqrt((S2.x-S1.x)^2+(S2.y-S1.y)^2)
	return R1+R2 > Dist
end

function BoundingCircleCollideWT(S1,S2,t1,t2)
	local Rad1 = math.sqrt((S1.image:getWidth()/2)^2+(S1.image:getHeight()/2)^2)-t1
	local Rad2 = math.sqrt((S2.image:getWidth()/2)^2+(S2.image:getHeight()/2)^2)-t2
	local Dist = math.sqrt((S2.x-S1.x)^2+(S2.y-S1.y)^2)
	return Rad1+Rad2 > Dist
end

--[[
This may seem silly but it extends the line so that no mater where the line is in a 800x600
screen. The line will always (unless the line is a point) reach the edge of the screen one way
(going from point 1 to point 2). You shouldn't need any more than this as even on the largest
screens I know of you would have to have a line of 3 pixels or less for it to matter.
Although, if you think it does matter then feel free to increase it. If you want to detect things off screen from
very far away then you may want to increase it.
@param:
   L-a table representing a line. must have x1,y1,x2,y2 keys as non-nil numerical values
--]]
function ExtendLine(L)
	local x = L.x1 - L.x2
	local y = L.y1 - L.y2
	L2 = {x1=L.x1,y1=L.y1,x2=L.x2+x*1000,y2=L.y2+y*1000}
	return L2
end


--[[
Standard box vs. box collision. Checks for bounding box collision between two sprites.
@param:
   S1-a table representing a sprite. must have x(number),y(number),image(LOVE image class)
   S2-same as S1
--]]
function BoundingBoxCollide(S1,S2)
	local left1=S1.x-S1.image:getWidth()/2
	local left2=S2.x-S2.image:getWidth()/2
	local right1=S1.x+S1.image:getWidth()/2
	local right2=S2.x+S2.image:getWidth()/2
	local top1=S1.y-S1.image:getHeight()/2
	local top2=S2.y-S2.image:getHeight()/2
	local bottom1=S1.y+S1.image:getHeight()/2
	local bottom2=S2.y+S2.image:getHeight()/2
	if bottom1 < top2 then return false end
    if top1 > bottom2 then return false end
    if right1 < left2 then return false end
    if left1 > right2 then return false end
	return true
end

--[[
Standard box vs. box collision but with tolerance so that you can shrink your bounding box for each sprite.
12 is a good tolerance for 64x64 images but it will take some messing with in most cases.
@param:
   S1-a table representing a sprite. must have x(number),y(number),image(LOVE image class).
   S2-same as S1.
   t1-number, tells how much to shrink bounding box by for S1.
   t2-same as t1 but instead for S2.
--]]
function BoundingBoxCollideWT(S1,S2,t1,t2)
	local left1=S1.x-S1.image:getWidth()/2+t1
	local left2=S2.x-S2.image:getWidth()/2+t2
	local right1=S1.x+S1.image:getWidth()/2-t1
	local right2=S2.x+S2.image:getWidth()/2-t2
	local top1=S1.y-S1.image:getHeight()/2+t1
	local top2=S2.y-S2.image:getHeight()/2+t2
	local bottom1=S1.y+S1.image:getHeight()/2-t1
	local bottom2=S2.y+S2.image:getHeight()/2-t2
	if bottom1 < top2 then return false end
    if top1 > bottom2 then return false end
    if right1 < left2 then return false end
    if left1 > right2 then return false end
	return true
end

--[[
This is mainly for BoxLineCollsion but you can use it for other stuff to
check for intersections between two lines.
@param:
   L1-a table representing a line. must have x1,y1,x2,y2 keys as non-nil numerical values.
   L2-same as L1.
--]]
function LineIntersets(L1,L2)
	local denom = ((L2.y2-L2.y1)*(L1.x2-L1.x1)-((L2.x2-L2.x1)*(L1.y2-L1.y1)))
	local numerator = ((L2.x2-L2.x1)*(L1.y1-L2.y1)-((L2.y2-L2.y1)*(L1.x1-L2.x1)))
	local numerator2 = ((L1.x2-L1.x1)*(L1.y1-L2.y1)-((L1.y2-L1.y1)*(L1.x1-L2.x1)))
	if denom == 0 then
		return false
	end
	local ua = numerator/denom
	local ub = numerator2/denom
	return (ua>=0 and ua <= 1 and ub >= 0 and ub <= 1)
end


--[[
Checks for intersections between a line and bounding box(of sprite).
@param:
   S-a table representing a sprite. must have x(number),y(number),image(LOVE image class).
   L-a table representing a line. must have x1,y1,x2,y2 keys as non-nil numerical values.
--]]
function BoxLineCollsion(S,L)
	local left=S.x-S.image:getWidth()/2+12
	local right=S.x+S.image:getWidth()/2-12
	local top=S.y-S.image:getHeight()/2+12
	local bottom=S.y+S.image:getHeight()/2-12
	if LineIntersets(NewLine(left,top,right,top),L) then return true
	elseif LineIntersets(NewLine(left,bottom,right,bottom),L) then return true
	elseif LineIntersets(NewLine(right,top,right,bottom),L) then return true
	--no need to check thrid as there is no way to go though 1 line and 1 line only
	end
	return false
end

--[[
Makes pixel map of image for pixel-perfect collision detection.
@param:
   imageFileName- string donating file path to image.
--]]
function NewCollisionMap( imageFileName )

    local imageData = love.image.newImageData( imageFileName )
    local width = imageData:getWidth()
    local height = imageData:getHeight()
	collisionMap = {}

    -- Build a collision map as a table of row tables that contains 1's and 0's.
    -- A 1 means the pixel at this position is non-transparent and 0 means it
    -- transparent.
    for y = 0, height do

        collisionMap[y] = {}

        for x = 0, width do

            -- Use -1 since getPixel() starts indexing at 0 not 1 like Lua.
            local r, g, b, a = imageData:getPixel( x, y )

            if a == 0 then
                collisionMap[y][x] = false
            else
                collisionMap[y][x] = true
            end

        end
    end

    return collisionMap

end


--[[
Checks for pixel perfect collision between two sprites.
@param:
   sprite1-a table representing a sprite. must have x(number),y(number),
         image(LOVE image class),collisionMap(class donated by function above)
   sprite2-same as sprite1
--]]
function PerPixelCollision( sprite1, sprite2 )

    -- First, we do a simple bounding box collision check. This will let
    -- us know if the two sprites overlap in any way.
    if not ( (sprite1.x + sprite1.image:getWidth() > sprite2.x) and
             (sprite1.x < sprite2.x + sprite2.image:getWidth()) and
             (sprite1.y + sprite1.image:getHeight() > sprite2.y) and
             (sprite1.y < sprite2.y + sprite2.image:getHeight()) ) then
        return false
    end

    -- If we made it this far, our two sprites definitely touch or overlap,
    -- but that doesn't mean that that we have an actual collision between
    -- two non-transparent pixels.

    -- By default, sprite1 scans sprite2 for a pixel collision per line, so
    -- if sprite1 is taller, swap the sprites around so the shorter one is
    -- scanning the taller one. This will result in less work in cases where
    -- they initially overlap but ultimately do not collide at the pixel level.
    if sprite1.image:getHeight() > sprite2.image:getHeight() then
       objTemp = sprite1
       sprite1 = sprite2
       sprite1 = objTemp
    end
    -- Loop through each row of sprite1's collision map and check it against
    -- sprite2's corresponding collision map row.
    for indexY = 0, sprite1.image:getHeight() do

        local screenY = math.floor( (sprite1.y + indexY) - 1  )

        if screenY > sprite2.y and screenY <= sprite2.y + sprite2.image:getHeight() then

            -- Some, or all, of the current row (Y) of sprite1's collision map overlaps
            -- sprite2's collision map. Calculate the start and end indices (X) for each
            -- row, so we can test this area of overlap for a collision of
            -- non-transparent pixels.

            local y1 = math.floor( indexY )
            local y2 = 1

            if screenY > sprite2.y then
                y2 = math.floor( screenY - sprite2.y )
            elseif screenY < sprite2.y then
                y2 = math.floor( sprite2.y - screenY )
            end

            local sprite1Index1 = 1
            local sprite1Index2 = sprite1.image:getWidth()
            local sprite2Index1 = 1
            local sprite2Index2 = sprite2.image:getWidth()

            if sprite1.x < sprite2.x then

               sprite1Index1 = math.floor( sprite2.x - sprite1.x ) + 1
               sprite1Index2 = sprite1.image:getWidth()

               sprite2Index1 = 1
               sprite2Index2 = math.floor( (sprite1.x + sprite1.image:getWidth()) - sprite2.x ) + 1

               -- If the sprites being tested are of different sizes it's possible
               -- for this index to get too big - so clamp it.
               if sprite2Index2 > sprite2.image:getWidth() then
                  sprite2Index2 = sprite2.image:getWidth()
               end

            elseif sprite1.x > sprite2.x then

               sprite1Index1 = 1
               sprite1Index2 = math.floor( (sprite2.x + sprite2.image:getWidth()) - sprite1.x ) + 1

               -- If the sprites being tested are of different sizes it's possible
               -- for this index to get too big - so clamp it.
               if sprite1Index2 > sprite1.image:getWidth() then
                  sprite1Index2 = sprite1.image:getWidth()
               end

               sprite2Index1 = math.floor( sprite1.x - sprite2.x ) + 1
               sprite2Index2 = sprite2.image:getWidth()

            else -- sprite1.x == sprite2.x

               -- If the two sprites have the same x position - the width of
               -- overlap is simply the shortest width.
               shortest = sprite1.image:getWidth()

               if sprite2.image:getWidth() < shortest then
                  shortest = sprite2.image:getWidth()
               end

               sprite1Index1 = 1
               sprite1Index2 = shortest

               sprite2Index1 = 1
               sprite2Index2 = shortest

            end

            local index1 = sprite1Index1
            local index2 = sprite2Index1

            while index1 < sprite1Index2 and index2 < sprite2Index2 do
                if sprite1.collisionMap[y1][index1] and sprite2.collisionMap[y2][index2] then
                    return true -- We have a collision of two non-transparent pixels!
                end

                index1 = index1 + 1
                index2 = index2 + 1

            end

        end

    end

    return false -- We do NOT have a collision of two non-transparent pixels.

end
